@@ -7,6 +7,7 @@
  */
 
 #include <linux/config.h>
+#include <linux/ctype.h>
 #include <linux/module.h>
 #include <linux/init.h>
 #include <linux/slab.h>
@@ -15,6 +16,7 @@
 #include <linux/blkdev.h>
 #include <linux/mempool.h>
 #include <linux/device-mapper.h>
+#include <linux/vmalloc.h>
 
 #include "dm-snapshot.h"
 #include "kcopyd.h"
@@ -33,16 +35,17 @@
  */
 #define WAKE_UP_PERCENT 5
 
-/*
- * Hard sector size used all over the kernel
- */
-#define SECTOR_SIZE 512
-
 /*
  * kcopyd priority of snapshot operations
  */
 #define SNAPSHOT_COPY_PRIORITY 2
 
+/*
+ * Each snapshot reserves this many pages for io
+ * FIXME: calculate this
+ */
+#define SNAPSHOT_PAGES 256
+
 struct pending_exception {
 	struct exception e;
 
@@ -205,7 +208,7 @@ static void unregister_snapshot(struct dm_snapshot *s)
  */
 static int init_exception_table(struct exception_table *et, uint32_t size)
 {
-	int i;
+	unsigned int i;
 
 	et->hash_mask = size - 1;
 	et->table = vcalloc(size, sizeof(struct list_head));
@@ -345,7 +348,7 @@ static inline uint32_t round_down(uint32_t n)
  */
 static int init_hash_tables(struct dm_snapshot *s)
 {
-	offset_t hash_size, cow_dev_size, origin_dev_size, max_buckets;
+	sector_t hash_size, cow_dev_size, origin_dev_size, max_buckets;
 
 	/*
 	 * Calculate based on the size of the original volume or
@@ -392,60 +395,61 @@ static inline ulong round_up(ulong n, ulong size)
 /*
  * Construct a snapshot mapping: <origin_dev> <COW-dev> <p/n> <chunk-size>
  */
-static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
-			int argc, char **argv, void **context)
+static int snapshot_ctr(struct dm_target *ti, unsigned int argc, char **argv)
 {
 	struct dm_snapshot *s;
 	unsigned long chunk_size;
 	int r = -EINVAL;
-	char *persistent;
+	char persistent;
 	char *origin_path;
 	char *cow_path;
 	char *value;
 	int blocksize;
 
 	if (argc < 4) {
-		*context = "dm-snapshot: requires exactly 4 arguments";
+		ti->error = "dm-snapshot: requires exactly 4 arguments";
 		r = -EINVAL;
-		goto bad;
+		goto bad1;
 	}
 
 	origin_path = argv[0];
 	cow_path = argv[1];
-	persistent = argv[2];
+	persistent = toupper(*argv[2]);
 
-	if ((*persistent & 0x5f) != 'P' && (*persistent & 0x5f) != 'N') {
-		*context = "Persistent flag is not P or N";
+	if (persistent != 'P' && persistent != 'N') {
+		ti->error = "Persistent flag is not P or N";
 		r = -EINVAL;
-		goto bad;
+		goto bad1;
 	}
 
 	chunk_size = simple_strtoul(argv[3], &value, 10);
 	if (chunk_size == 0 || value == NULL) {
-		*context = "Invalid chunk size";
+		ti->error = "Invalid chunk size";
 		r = -EINVAL;
-		goto bad;
+		goto bad1;
 	}
 
 	s = kmalloc(sizeof(*s), GFP_KERNEL);
 	if (s == NULL) {
-		*context = "Cannot allocate snapshot context private structure";
+		ti->error = "Cannot allocate snapshot context private "
+		    "structure";
 		r = -ENOMEM;
-		goto bad;
+		goto bad1;
 	}
 
-	r = dm_table_get_device(t, origin_path, 0, 0, FMODE_READ, &s->origin);
+	r = dm_get_device(ti, origin_path, 0, ti->len, FMODE_READ, &s->origin);
 	if (r) {
-		*context = "Cannot get origin device";
-		goto bad_free;
+		ti->error = "Cannot get origin device";
+		goto bad2;
 	}
 
-	r = dm_table_get_device(t, cow_path, 0, 0,
-				FMODE_READ | FMODE_WRITE, &s->cow);
+	/* FIXME: get cow length */
+	r = dm_get_device(ti, cow_path, 0, 0,
+			  FMODE_READ | FMODE_WRITE, &s->cow);
 	if (r) {
-		dm_table_put_device(t, s->origin);
-		*context = "Cannot get COW device";
-		goto bad_free;
+		dm_put_device(ti, s->origin);
+		ti->error = "Cannot get COW device";
+		goto bad2;
 	}
 
 	/*
@@ -457,43 +461,44 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 	/* Validate the chunk size against the device block size */
 	blocksize = get_hardsect_size(s->cow->dev);
 	if (chunk_size % (blocksize / SECTOR_SIZE)) {
-		*context = "Chunk size is not a multiple of device blocksize";
+		ti->error = "Chunk size is not a multiple of device blocksize";
 		r = -EINVAL;
-		goto bad_putdev;
+		goto bad3;
 	}
 
 	/* Check the sizes are small enough to fit in one kiovec */
 	if (chunk_size > KIO_MAX_SECTORS) {
-		*context = "Chunk size is too big";
+		ti->error = "Chunk size is too big";
 		r = -EINVAL;
-		goto bad_putdev;
+		goto bad3;
 	}
 
 	/* Check chunk_size is a power of 2 */
 	if (chunk_size & (chunk_size - 1)) {
-		*context = "Chunk size is not a power of 2";
+		ti->error = "Chunk size is not a power of 2";
 		r = -EINVAL;
-		goto bad_putdev;
+		goto bad3;
 	}
 
 	s->chunk_size = chunk_size;
 	s->chunk_mask = chunk_size - 1;
-	s->type = *persistent;
+	s->type = persistent;
 	for (s->chunk_shift = 0; chunk_size;
 	     s->chunk_shift++, chunk_size >>= 1)
 		;
 	s->chunk_shift--;
 
 	s->valid = 1;
+	s->have_metadata = 0;
 	s->last_percent = 0;
-	s->table = t;
 	init_rwsem(&s->lock);
+	s->table = ti->table;
 
 	/* Allocate hash table for COW data */
 	if (init_hash_tables(s)) {
-		*context = "Unable to allocate hash table space";
+		ti->error = "Unable to allocate hash table space";
 		r = -ENOMEM;
-		goto bad_putdev;
+		goto bad3;
 	}
 
 	/*
@@ -502,61 +507,62 @@ static int snapshot_ctr(struct dm_table *t, offset_t b, offset_t l,
 	 */
 	s->store.snap = s;
 
-	if ((*persistent & 0x5f) == 'P')
+	if (persistent == 'P')
 		r = dm_create_persistent(&s->store, s->chunk_size);
 	else
-		r = dm_create_transient(&s->store, s, blocksize, context);
+		r = dm_create_transient(&s->store, s, blocksize);
 
 	if (r) {
-		*context = "Couldn't create exception store";
+		ti->error = "Couldn't create exception store";
 		r = -EINVAL;
-		goto bad_free1;
+		goto bad4;
+	}
+
+	r = kcopyd_client_create(SNAPSHOT_PAGES, &s->kcopyd_client);
+	if (r) {
+		ti->error = "Could not create kcopyd client";
+		goto bad5;
 	}
 
 	/* Flush IO to the origin device */
-#if LVM_VFS_ENHANCEMENT
-	fsync_dev_lockfs(s->origin->dev);
-#else
 	fsync_dev(s->origin->dev);
-#endif
 
 	/* Add snapshot to the list of snapshots for this origin */
 	if (register_snapshot(s)) {
 		r = -EINVAL;
-		*context = "Cannot register snapshot origin";
-		goto bad_free2;
+		ti->error = "Cannot register snapshot origin";
+		goto bad6;
 	}
-#if LVM_VFS_ENHANCEMENT
-	unlockfs(s->origin->dev);
-#endif
-	kcopyd_inc_client_count();
 
-	*context = s;
+	ti->private = s;
 	return 0;
 
-      bad_free2:
+ bad6:
+	kcopyd_client_destroy(s->kcopyd_client);
+
+ bad5:
 	s->store.destroy(&s->store);
 
-      bad_free1:
+ bad4:
 	exit_exception_table(&s->pending, pending_cache);
 	exit_exception_table(&s->complete, exception_cache);
 
-      bad_putdev:
-	dm_table_put_device(t, s->cow);
-	dm_table_put_device(t, s->origin);
+ bad3:
+	dm_put_device(ti, s->cow);
+	dm_put_device(ti, s->origin);
 
-      bad_free:
+ bad2:
 	kfree(s);
 
-      bad:
+ bad1:
 	return r;
 }
 
-static void snapshot_dtr(struct dm_table *t, void *context)
+static void snapshot_dtr(struct dm_target *ti)
 {
-	struct dm_snapshot *s = (struct dm_snapshot *) context;
+	struct dm_snapshot *s = (struct dm_snapshot *) ti->private;
 
-	dm_table_event(s->table);
+	dm_table_event(ti->table);
 
 	unregister_snapshot(s);
 
@@ -566,11 +572,10 @@ static void snapshot_dtr(struct dm_table *t, void *context)
 	/* Deallocate memory used */
 	s->store.destroy(&s->store);
 
-	dm_table_put_device(t, s->origin);
-	dm_table_put_device(t, s->cow);
+	dm_put_device(ti, s->origin);
+	dm_put_device(ti, s->cow);
+	kcopyd_client_destroy(s->kcopyd_client);
 	kfree(s);
-
-	kcopyd_dec_client_count();
 }
 
 /*
@@ -582,6 +587,17 @@ static void queue_buffer(struct buffer_head **queue, struct buffer_head *bh)
 	*queue = bh;
 }
 
+/*
+ * FIXME: inefficient.
+ */
+static void queue_buffers(struct buffer_head **queue, struct buffer_head *bhs)
+{
+	while (*queue)
+		queue = &((*queue)->b_reqnext);
+
+	*queue = bhs;
+}
+
 /*
  * Flush a list of buffers.
  */
@@ -616,31 +632,54 @@ static void error_buffers(struct buffer_head *bh)
 	}
 }
 
+static struct buffer_head *__flush_bhs(struct pending_exception *pe)
+{
+	struct pending_exception *sibling;
+
+	if (list_empty(&pe->siblings))
+		return pe->origin_bhs;
+
+	sibling = list_entry(pe->siblings.next,
+			     struct pending_exception, siblings);
+
+	list_del(&pe->siblings);
+
+	/* FIXME: I think there's a race on SMP machines here, add spin lock */
+	queue_buffers(&sibling->origin_bhs, pe->origin_bhs);
+
+	return NULL;
+}
+
 static void pending_complete(struct pending_exception *pe, int success)
 {
 	struct exception *e;
 	struct dm_snapshot *s = pe->snap;
+	struct buffer_head *flush = NULL;
 
 	if (success) {
 		e = alloc_exception();
 		if (!e) {
-			printk("Unable to allocate exception.");
+			DMWARN("Unable to allocate exception.");
 			down_write(&s->lock);
 			s->store.drop_snapshot(&s->store);
 			s->valid = 0;
+			flush = __flush_bhs(pe);
 			up_write(&s->lock);
-			return;
+
+			error_buffers(pe->snapshot_bhs);
+			goto out;
 		}
 
 		/*
 		 * Add a proper exception, and remove the
-		 * inflight exception from the list.
+		 * in-flight exception from the list.
 		 */
 		down_write(&s->lock);
 
 		memcpy(e, &pe->e, sizeof(*e));
 		insert_exception(&s->complete, e);
 		remove_exception(&pe->e);
+		flush = __flush_bhs(pe);
 
 		/* Submit any pending write BHs */
 		up_write(&s->lock);
@@ -649,8 +688,13 @@ static void pending_complete(struct pending_exception *pe, int success)
 		DMDEBUG("Exception completed successfully.");
 
 		/* Notify any interested parties */
-		if (s->store.percent_full) {
-			int pc = s->store.percent_full(&s->store);
+		if (s->store.fraction_full) {
+			sector_t numerator, denominator;
+			int pc;
+
+			s->store.fraction_full(&s->store, &numerator,
+					       &denominator);
+			pc = numerator * 100 / denominator;
 
 			if (pc >= s->last_percent + WAKE_UP_PERCENT) {
 				dm_table_event(s->table);
@@ -660,12 +704,13 @@ static void pending_complete(struct pending_exception *pe, int success)
 
 	} else {
 		/* Read/write error - snapshot is unusable */
-		DMERR("Error reading/writing snapshot");
-
 		down_write(&s->lock);
+		if (s->valid)
+			DMERR("Error reading/writing snapshot");
 		s->store.drop_snapshot(&s->store);
 		s->valid = 0;
 		remove_exception(&pe->e);
+		flush = __flush_bhs(pe);
 		up_write(&s->lock);
 
 		error_buffers(pe->snapshot_bhs);
@@ -674,10 +719,9 @@ static void pending_complete(struct pending_exception *pe, int success)
 		DMDEBUG("Exception failed.");
 	}
 
-	if (list_empty(&pe->siblings))
-		flush_buffers(pe->origin_bhs);
-	else
-		list_del(&pe->siblings);
+ out:
+	if (flush)
+		flush_buffers(flush);
 
 	free_pending_exception(pe);
 }
@@ -692,12 +736,12 @@ static void commit_callback(void *context, int success)
  * Called when the copy I/O has finished.  kcopyd actually runs
  * this code so don't block.
  */
-static void copy_callback(int err, void *context)
+static void copy_callback(int read_err, unsigned int write_err, void *context)
 {
 	struct pending_exception *pe = (struct pending_exception *) context;
 	struct dm_snapshot *s = pe->snap;
 
-	if (err)
+	if (read_err || write_err)
 		pending_complete(pe, 0);
 
 	else
@@ -712,21 +756,31 @@ static void copy_callback(int err, void *context)
 static inline void start_copy(struct pending_exception *pe)
 {
 	struct dm_snapshot *s = pe->snap;
-	struct kcopyd_region src, dest;
+	struct io_region src, dest;
+	kdev_t dev = s->origin->dev;
+	int *sizes = blk_size[major(dev)];
+	sector_t dev_size = (sector_t) -1;
 
-	src.dev = s->origin->dev;
+	if (pe->started)
+		return;
+
+	/* this is protected by snap->lock */
+	pe->started = 1;
+
+	if (sizes && sizes[minor(dev)])
+		dev_size = sizes[minor(dev)] << 1;
+
+	src.dev = dev;
 	src.sector = chunk_to_sector(s, pe->e.old_chunk);
-	src.count = s->chunk_size;
+	src.count = min(s->chunk_size, dev_size - src.sector);
 
 	dest.dev = s->cow->dev;
 	dest.sector = chunk_to_sector(s, pe->e.new_chunk);
-	dest.count = s->chunk_size;
+	dest.count = src.count;
 
-	if (!pe->started) {
-		/* Hand over to kcopyd */
-		kcopyd_copy(&src, &dest, copy_callback, pe);
-		pe->started = 1;
-	}
+	/* Hand over to kcopyd */
+	kcopyd_copy(s->kcopyd_client,
+		    &src, 1, &dest, 0, copy_callback, pe);
 }
 
 /*
@@ -752,11 +806,6 @@ static struct pending_exception *find_pending_exception(struct dm_snapshot *s,
 	} else {
 		/* Create a new pending exception */
 		pe = alloc_pending_exception();
-		if (!pe) {
-			DMWARN("Couldn't allocate pending exception.");
-			return NULL;
-		}
-
 		pe->e.old_chunk = chunk;
 		pe->origin_bhs = pe->snapshot_bhs = NULL;
 		INIT_LIST_HEAD(&pe->siblings);
@@ -783,10 +832,11 @@ static inline void remap_exception(struct dm_snapshot *s, struct exception *e,
 	    (bh->b_rsector & s->chunk_mask);
 }
 
-static int snapshot_map(struct buffer_head *bh, int rw, void *context)
+static int snapshot_map(struct dm_target *ti, struct buffer_head *bh, int rw,
+			union map_info *map_context)
 {
 	struct exception *e;
-	struct dm_snapshot *s = (struct dm_snapshot *) context;
+	struct dm_snapshot *s = (struct dm_snapshot *) ti->private;
 	int r = 1;
 	chunk_t chunk;
 	struct pending_exception *pe;
@@ -817,11 +867,13 @@ static int snapshot_map(struct buffer_head *bh, int rw, void *context)
 			if (!pe) {
 				s->store.drop_snapshot(&s->store);
 				s->valid = 0;
+				r = -EIO;
+			} else {
+				remap_exception(s, &pe->e, bh);
+				queue_buffer(&pe->snapshot_bhs, bh);
+				start_copy(pe);
+				r = 0;
 			}
-
-			queue_buffer(&pe->snapshot_bhs, bh);
-			start_copy(pe);
-			r = 0;
 		}
 
 		up_write(&s->lock);
@@ -850,6 +902,67 @@ static int snapshot_map(struct buffer_head *bh, int rw, void *context)
 	return r;
 }
 
+void snapshot_resume(struct dm_target *ti)
+{
+	struct dm_snapshot *s = (struct dm_snapshot *) ti->private;
+
+	if (s->have_metadata)
+		return;
+
+	if (s->store.read_metadata(&s->store)) {
+		down_write(&s->lock);
+		s->valid = 0;
+		up_write(&s->lock);
+	}
+
+	s->have_metadata = 1;
+}
+
+static int snapshot_status(struct dm_target *ti, status_type_t type,
+			   char *result, unsigned int maxlen)
+{
+	struct dm_snapshot *snap = (struct dm_snapshot *) ti->private;
+	char cow[16];
+	char org[16];
+
+	switch (type) {
+	case STATUSTYPE_INFO:
+		if (!snap->valid)
+			snprintf(result, maxlen, "Invalid");
+		else {
+			if (snap->store.fraction_full) {
+				sector_t numerator, denominator;
+				snap->store.fraction_full(&snap->store,
+							  &numerator,
+							  &denominator);
+				snprintf(result, maxlen,
+					 SECTOR_FORMAT "/" SECTOR_FORMAT,
+					 numerator, denominator);
+			}
+			else
+				snprintf(result, maxlen, "Unknown");
+		}
+		break;
+
+	case STATUSTYPE_TABLE:
+		/*
+		 * kdevname returns a static pointer so we need
+		 * to make private copies if the output is to
+		 * make sense.
+		 */
+		strncpy(cow, dm_kdevname(snap->cow->dev), sizeof(cow));
+		strncpy(org, dm_kdevname(snap->origin->dev), sizeof(org));
+		snprintf(result, maxlen, "%s %s %c %ld", org, cow,
+			 snap->type, snap->chunk_size);
+		break;
+	}
+
+	return 0;
+}
+
+/*-----------------------------------------------------------------
+ * Origin methods
+ *---------------------------------------------------------------*/
 static void list_merge(struct list_head *l1, struct list_head *l2)
 {
 	struct list_head *l1_n, *l2_p;
@@ -866,7 +979,7 @@ static void list_merge(struct list_head *l1, struct list_head *l2)
 
 static int __origin_write(struct list_head *snapshots, struct buffer_head *bh)
 {
-	int r = 1;
+	int r = 1, first = 1;
 	struct list_head *sl;
 	struct dm_snapshot *snap;
 	struct exception *e;
@@ -921,9 +1034,11 @@ static int __origin_write(struct list_head *snapshots, struct buffer_head *bh)
 		pe = last;
 		do {
 			down_write(&pe->snap->lock);
-			queue_buffer(&pe->origin_bhs, bh);
+			if (first)
+				queue_buffer(&pe->origin_bhs, bh);
 			start_copy(pe);
 			up_write(&pe->snap->lock);
+			first = 0;
 			pe = list_entry(pe->siblings.next,
 					struct pending_exception, siblings);
 
@@ -933,43 +1048,6 @@ static int __origin_write(struct list_head *snapshots, struct buffer_head *bh)
 	return r;
 }
 
-static int snapshot_status(status_type_t type, char *result,
-			   int maxlen, void *context)
-{
-	struct dm_snapshot *snap = (struct dm_snapshot *) context;
-	char cow[16];
-	char org[16];
-
-	switch (type) {
-	case STATUSTYPE_INFO:
-		if (!snap->valid)
-			snprintf(result, maxlen, "Invalid");
-		else {
-			if (snap->store.percent_full)
-				snprintf(result, maxlen, "%d%%",
-					 snap->store.percent_full(&snap->
-								  store));
-			else
-				snprintf(result, maxlen, "Unknown");
-		}
-		break;
-
-	case STATUSTYPE_TABLE:
-		/*
-		 * kdevname returns a static pointer so we need
-		 * to make private copies if the output is to
-		 * make sense.
-		 */
-		strncpy(cow, kdevname(snap->cow->dev), sizeof(cow));
-		strncpy(org, kdevname(snap->origin->dev), sizeof(org));
-		snprintf(result, maxlen, "%s %s %c %ld", org, cow,
-			 snap->type, snap->chunk_size);
-		break;
-	}
-
-	return 0;
-}
-
 /*
  * Called on a write from the origin driver.
  */
@@ -998,47 +1076,47 @@ int do_origin(struct dm_dev *origin, struct buffer_head *bh)
  * The context for an origin is merely a 'struct dm_dev *'
  * pointing to the real device.
  */
-static int origin_ctr(struct dm_table *t, offset_t b, offset_t l,
-		      int argc, char **argv, void **context)
+static int origin_ctr(struct dm_target *ti, unsigned int argc, char **argv)
 {
 	int r;
 	struct dm_dev *dev;
 
 	if (argc != 1) {
-		*context = "dm-origin: incorrect number of arguments";
+		ti->error = "dm-origin: incorrect number of arguments";
 		return -EINVAL;
 	}
 
-	r = dm_table_get_device(t, argv[0], 0, l, t->mode, &dev);
+	r = dm_get_device(ti, argv[0], 0, ti->len,
+			  dm_table_get_mode(ti->table), &dev);
 	if (r) {
-		*context = "Cannot get target device";
+		ti->error = "Cannot get target device";
 		return r;
 	}
 
-	*context = dev;
-
+	ti->private = dev;
 	return 0;
 }
 
-static void origin_dtr(struct dm_table *t, void *c)
+static void origin_dtr(struct dm_target *ti)
 {
-	struct dm_dev *dev = (struct dm_dev *) c;
-	dm_table_put_device(t, dev);
+	struct dm_dev *dev = (struct dm_dev *) ti->private;
+	dm_put_device(ti, dev);
 }
 
-static int origin_map(struct buffer_head *bh, int rw, void *context)
+static int origin_map(struct dm_target *ti, struct buffer_head *bh, int rw,
+		      union map_info *map_context)
 {
-	struct dm_dev *dev = (struct dm_dev *) context;
+	struct dm_dev *dev = (struct dm_dev *) ti->private;
 	bh->b_rdev = dev->dev;
 
 	/* Only tell snapshots if this is a write */
 	return (rw == WRITE) ? do_origin(dev, bh) : 1;
 }
 
-static int origin_status(status_type_t type, char *result,
-			 int maxlen, void *context)
+static int origin_status(struct dm_target *ti, status_type_t type, char *result,
+			 unsigned int maxlen)
 {
-	struct dm_dev *dev = (struct dm_dev *) context;
+	struct dm_dev *dev = (struct dm_dev *) ti->private;
 
 	switch (type) {
 	case STATUSTYPE_INFO:
@@ -1046,7 +1124,7 @@ static int origin_status(status_type_t type, char *result,
 		break;
 
 	case STATUSTYPE_TABLE:
-		snprintf(result, maxlen, "%s", kdevname(dev->dev));
+		snprintf(result, maxlen, "%s", dm_kdevname(dev->dev));
 		break;
 	}
 
@@ -1060,7 +1138,6 @@ static struct target_type origin_target = {
 	dtr:	origin_dtr,
 	map:	origin_map,
 	status:	origin_status,
-	err:	NULL
 };
 
 static struct target_type snapshot_target = {
@@ -1069,8 +1146,8 @@ static struct target_type snapshot_target = {
 	ctr:	snapshot_ctr,
 	dtr:	snapshot_dtr,
 	map:	snapshot_map,
+	resume: snapshot_resume,
 	status:	snapshot_status,
-	err:	NULL
 };
 
 int __init dm_snapshot_init(void)
@@ -1156,14 +1233,3 @@ void dm_snapshot_exit(void)
 	kmem_cache_destroy(pending_cache);
 	kmem_cache_destroy(exception_cache);
 }
-
-/*
- * Overrides for Emacs so that we follow Linus's tabbing style.
- * Emacs will notice this stuff at the end of the file and automatically
- * adjust the settings for this buffer only.  This must remain at the end
- * of the file.
- * ---------------------------------------------------------------------------
- * Local variables:
- * c-file-style: "linux"
- * End:
- */
