@@ -30,6 +30,7 @@ struct hash_cell {
 	char *name;
 	char *uuid;
 	struct mapped_device *md;
+	struct dm_table *new_map;
 
 	/* I hate devfs */
 	devfs_handle_t devfs_entry;
@@ -44,7 +45,7 @@ static devfs_handle_t _dev_dir;
 void dm_hash_remove_all(void);
 
 /*
- * Guards access to all three tables.
+ * Guards access to both hash tables.
  */
 static DECLARE_RWSEM(_hash_lock);
 
@@ -95,7 +96,7 @@ static struct hash_cell *__get_name_cell(const char *str)
 	struct hash_cell *hc;
 	unsigned int h = hash_str(str);
 
-	list_for_each(tmp, _name_buckets + h) {
+	list_for_each (tmp, _name_buckets + h) {
 		hc = list_entry(tmp, struct hash_cell, name_list);
 		if (!strcmp(hc->name, str))
 			return hc;
@@ -110,7 +111,7 @@ static struct hash_cell *__get_uuid_cell(const char *str)
 	struct hash_cell *hc;
 	unsigned int h = hash_str(str);
 
-	list_for_each(tmp, _uuid_buckets + h) {
+	list_for_each (tmp, _uuid_buckets + h) {
 		hc = list_entry(tmp, struct hash_cell, uuid_list);
 		if (!strcmp(hc->uuid, str))
 			return hc;
@@ -160,6 +161,7 @@ static struct hash_cell *alloc_cell(const char *name, const char *uuid,
 	INIT_LIST_HEAD(&hc->name_list);
 	INIT_LIST_HEAD(&hc->uuid_list);
 	hc->md = md;
+	hc->new_map = NULL;
 	return hc;
 }
 
@@ -210,7 +212,7 @@ int dm_hash_insert(const char *name, const char *uuid, struct mapped_device *md)
 		return -ENOMEM;
 
 	/*
-	 * Insert the cell into all three hash tables.
+	 * Insert the cell into both hash tables.
 	 */
 	down_write(&_hash_lock);
 	if (__get_name_cell(name))
@@ -244,6 +246,9 @@ void __hash_remove(struct hash_cell *hc)
 	list_del(&hc->name_list);
 	unregister_with_devfs(hc);
 	dm_put(hc->md);
+	if (hc->new_map)
+		dm_table_put(hc->new_map);
+	free_cell(hc);
 }
 
 void dm_hash_remove_all(void)
@@ -254,7 +259,7 @@ void dm_hash_remove_all(void)
 
 	down_write(&_hash_lock);
 	for (i = 0; i < NUM_BUCKETS; i++) {
-		list_for_each_safe(tmp, n, _name_buckets + i) {
+		list_for_each_safe (tmp, n, _name_buckets + i) {
 			hc = list_entry(tmp, struct hash_cell, name_list);
 			__hash_remove(hc);
 		}
@@ -284,6 +289,7 @@ int dm_hash_rename(const char *old, const char *new)
 		DMWARN("asked to rename to an already existing name %s -> %s",
 		       old, new);
 		up_write(&_hash_lock);
+		kfree(new_name);
 		return -EBUSY;
 	}
 
@@ -295,6 +301,7 @@ int dm_hash_rename(const char *old, const char *new)
 		DMWARN("asked to rename a non existent device %s -> %s",
 		       old, new);
 		up_write(&_hash_lock);
+		kfree(new_name);
 		return -ENXIO;
 	}
 
@@ -315,176 +322,201 @@ int dm_hash_rename(const char *old, const char *new)
 	return 0;
 }
 
-
 /*-----------------------------------------------------------------
  * Implementation of the ioctl commands
  *---------------------------------------------------------------*/
-
 /*
  * All the ioctl commands get dispatched to functions with this
  * prototype.
  */
-typedef int (*ioctl_fn)(struct dm_ioctl *param, struct dm_ioctl *user);
+typedef int (*ioctl_fn)(struct dm_ioctl *param, size_t param_size);
+
+static int remove_all(struct dm_ioctl *param, size_t param_size)
+{
+	dm_hash_remove_all();
+	param->data_size = 0;
+	return 0;
+}
 
 /*
- * Check a string doesn't overrun the chunk of
- * memory we copied from userland.
+ * Round up the ptr to an 8-byte boundary.
  */
-static int valid_str(char *str, void *begin, void *end)
+#define ALIGN_MASK 7
+static inline void *align_ptr(void *ptr)
 {
-	while (((void *) str >= begin) && ((void *) str < end))
-		if (!*str++)
-			return 0;
-
-	return -EINVAL;
+	return (void *) (((size_t) (ptr + ALIGN_MASK)) & ~ALIGN_MASK);
 }
 
-static int next_target(struct dm_target_spec *last, uint32_t next,
-		       void *begin, void *end,
-		       struct dm_target_spec **spec, char **params)
+/*
+ * Retrieves the data payload buffer from an already allocated
+ * struct dm_ioctl.
+ */
+static void *get_result_buffer(struct dm_ioctl *param, size_t param_size,
+			       size_t *len)
 {
-	*spec = (struct dm_target_spec *)
-	    ((unsigned char *) last + next);
-	*params = (char *) (*spec + 1);
+	param->data_start = align_ptr(param + 1) - (void *) param;
 
-	if (*spec < (last + 1) || ((void *) *spec > end))
-		return -EINVAL;
+	if (param->data_start < param_size)
+		*len = param_size - param->data_start;
+	else
+		*len = 0;
 
-	return valid_str(*params, begin, end);
+	return ((void *) param) + param->data_start;
 }
 
-static int populate_table(struct dm_table *table, struct dm_ioctl *args)
+static int list_devices(struct dm_ioctl *param, size_t param_size)
 {
-	int i = 0, r, first = 1;
-	struct dm_target_spec *spec;
-	char *params;
-	void *begin, *end;
+	unsigned int i;
+	struct hash_cell *hc;
+	size_t len, needed = 0;
+	struct dm_name_list *nl, *old_nl = NULL;
 
-	if (!args->target_count) {
-		DMWARN("populate_table: no targets specified");
-		return -EINVAL;
-	}
+	down_write(&_hash_lock);
 
-	begin = (void *) args;
-	end = begin + args->data_size;
-
-	for (i = 0; i < args->target_count; i++) {
-
-		if (first)
-			r = next_target((struct dm_target_spec *) args,
-					args->data_start,
-					begin, end, &spec, &params);
-		else
-			r = next_target(spec, spec->next, begin, end,
-					&spec, &params);
-
-		if (r) {
-			DMWARN("unable to find target");
-			return -EINVAL;
+	/*
+	 * Loop through all the devices working out how much
+	 * space we need.
+	 */
+	for (i = 0; i < NUM_BUCKETS; i++) {
+		list_for_each_entry (hc, _name_buckets + i, name_list) {
+			needed += sizeof(struct dm_name_list);
+			needed += strlen(hc->name);
+			needed += ALIGN_MASK;
 		}
-
-		r = dm_table_add_target(table, spec->target_type,
-					spec->sector_start, spec->length,
-					params);
-		if (r) {
-			DMWARN("error adding target to table");
-			return -EINVAL;
-		}
-
-		first = 0;
-	}
-
-	return dm_table_complete(table);
-}
-
-/*
- * Round up the ptr to the next 'align' boundary.  Obviously
- * 'align' must be a power of 2.
- */
-static inline void *align_ptr(void *ptr, unsigned int align)
-{
-	align--;
-	return (void *) (((unsigned long) (ptr + align)) & ~align);
-}
-
-/*
- * Copies a dm_ioctl and an optional additional payload to
- * userland.
- */
-static int results_to_user(struct dm_ioctl *user, struct dm_ioctl *param,
-			   void *data, uint32_t len)
-{
-	int r;
-	void *ptr = NULL;
-
-	if (data) {
-		ptr = align_ptr(user + 1, sizeof(unsigned long));
-		param->data_start = ptr - (void *) user;
 	}
 
 	/*
-	 * The version number has already been filled in, so we
-	 * just copy later fields.
+	 * Grab our output buffer.
 	 */
-	r = copy_to_user(&user->data_size, &param->data_size,
-			 sizeof(*param) - sizeof(param->version));
-	if (r)
-		return -EFAULT;
+	nl = get_result_buffer(param, param_size, &len);
+	if (len < needed) {
+		param->flags |= DM_BUFFER_FULL_FLAG;
+		goto out;
+	}
+	param->data_size = param->data_start + needed;
 
-	if (data) {
-		if (param->data_start + len > param->data_size)
-			return -ENOSPC;
+	nl->dev = 0;	/* Flags no data */
 
-		if (copy_to_user(ptr, data, len))
-			r = -EFAULT;
+	/*
+	 * Now loop through filling out the names.
+	 */
+	for (i = 0; i < NUM_BUCKETS; i++) {
+		list_for_each_entry (hc, _name_buckets + i, name_list) {
+			if (old_nl)
+				old_nl->next = (uint32_t) ((void *) nl -
+							   (void *) old_nl);
+
+			nl->dev = dm_kdev(hc->md);
+			nl->next = 0;
+			strcpy(nl->name, hc->name);
+
+			old_nl = nl;
+			nl = align_ptr(((void *) ++nl) + strlen(hc->name) + 1);
+		}
 	}
 
-	return r;
+ out:
+	up_write(&_hash_lock);
+	return 0;
+}
+
+static int check_name(const char *name)
+{
+	if (strchr(name, '/')) {
+		DMWARN("invalid device name");
+		return -EINVAL;
+	}
+
+	return 0;
 }
 
 /*
  * Fills in a dm_ioctl structure, ready for sending back to
  * userland.
  */
-static int __info(struct mapped_device *md, struct dm_ioctl *param)
+static int __dev_status(struct mapped_device *md, struct dm_ioctl *param)
 {
 	kdev_t dev = dm_kdev(md);
 	struct dm_table *table;
 	struct block_device *bdev;
 
-	param->flags = DM_EXISTS_FLAG;
+	param->flags &= ~(DM_SUSPEND_FLAG | DM_READONLY_FLAG |
+			  DM_ACTIVE_PRESENT_FLAG);
+
 	if (dm_suspended(md))
 		param->flags |= DM_SUSPEND_FLAG;
 
 	param->dev = kdev_t_to_nr(dev);
-	bdev = bdget(param->dev);
-	if (!bdev)
-		return -ENXIO;
-
-	param->open_count = bdev->bd_openers;
-	bdput(bdev);
 
 	if (is_read_only(dev))
 		param->flags |= DM_READONLY_FLAG;
 
+	param->event_nr = dm_get_event_nr(md);
+
 	table = dm_get_table(md);
-	param->target_count = dm_table_get_num_targets(table);
-	dm_table_put(table);
+	if (table) {
+		param->flags |= DM_ACTIVE_PRESENT_FLAG;
+		param->target_count = dm_table_get_num_targets(table);
+		dm_table_put(table);
+	} else
+		param->target_count = 0;
+
+	bdev = bdget(param->dev);
+	if (!bdev)
+		return -ENXIO;
+	param->open_count = bdev->bd_openers;
+	bdput(bdev);
 
 	return 0;
 }
 
+static int dev_create(struct dm_ioctl *param, size_t param_size)
+{
+	int r;
+	kdev_t dev = 0;
+	struct mapped_device *md;
+
+	r = check_name(param->name);
+	if (r)
+		return r;
+
+	if (param->flags & DM_PERSISTENT_DEV_FLAG)
+		dev = to_kdev_t(param->dev);
+
+	r = dm_create(dev, &md);
+	if (r)
+		return r;
+
+	r = dm_hash_insert(param->name, *param->uuid ? param->uuid : NULL, md);
+	if (r) {
+		dm_put(md);
+		return r;
+	}
+
+	param->flags &= ~DM_INACTIVE_PRESENT_FLAG;
+
+	r = __dev_status(md, param);
+	dm_put(md);
+
+	return r;
+}
+
 /*
  * Always use UUID for lookups if it's present, otherwise use name.
  */
+static inline struct hash_cell *__find_device_hash_cell(struct dm_ioctl *param)
+{
+	return *param->uuid ?
+	    __get_uuid_cell(param->uuid) : __get_name_cell(param->name);
+}
+
 static inline struct mapped_device *find_device(struct dm_ioctl *param)
 {
 	struct hash_cell *hc;
 	struct mapped_device *md = NULL;
 
 	down_read(&_hash_lock);
-	hc = *param->uuid ? __get_uuid_cell(param->uuid) :
-	    __get_name_cell(param->name);
+	hc = __find_device_hash_cell(param);
 	if (hc) {
 		md = hc->md;
 
@@ -498,6 +530,11 @@ static inline struct mapped_device *find_device(struct dm_ioctl *param)
 		else
 			param->uuid[0] = '\0';
 
+		if (hc->new_map)
+			param->flags |= DM_INACTIVE_PRESENT_FLAG;
+		else
+			param->flags &= ~DM_INACTIVE_PRESENT_FLAG;
+
 		dm_get(md);
 	}
 	up_read(&_hash_lock);
@@ -505,36 +542,160 @@ static inline struct mapped_device *find_device(struct dm_ioctl *param)
 	return md;
 }
 
-#define ALIGNMENT sizeof(int)
-static void *_align(void *ptr, unsigned int a)
+static int dev_remove(struct dm_ioctl *param, size_t param_size)
 {
-	register unsigned long align = --a;
+	struct hash_cell *hc;
 
-	return (void *) (((unsigned long) ptr + align) & ~align);
+	down_write(&_hash_lock);
+	hc = __find_device_hash_cell(param);
+
+	if (!hc) {
+		DMWARN("device doesn't appear to be in the dev hash table.");
+		up_write(&_hash_lock);
+		return -ENXIO;
+	}
+
+	__hash_remove(hc);
+	up_write(&_hash_lock);
+	param->data_size = 0;
+	return 0;
+}
+
+/*
+ * Check a string doesn't overrun the chunk of
+ * memory we copied from userland.
+ */
+static int invalid_str(char *str, void *end)
+{
+	while ((void *) str < end)
+		if (!*str++)
+			return 0;
+
+	return -EINVAL;
+}
+
+static int dev_rename(struct dm_ioctl *param, size_t param_size)
+{
+	int r;
+	char *new_name = (char *) param + param->data_start;
+
+	if (new_name < (char *) (param + 1) ||
+	    invalid_str(new_name, (void *) param + param_size)) {
+		DMWARN("Invalid new logical volume name supplied.");
+		return -EINVAL;
+	}
+
+	r = check_name(new_name);
+	if (r)
+		return r;
+
+	param->data_size = 0;
+	return dm_hash_rename(param->name, new_name);
+}
+
+static int do_suspend(struct dm_ioctl *param)
+{
+	int r = 0;
+	struct mapped_device *md;
+
+	md = find_device(param);
+	if (!md)
+		return -ENXIO;
+
+	if (!dm_suspended(md))
+		r = dm_suspend(md);
+
+	if (!r)
+		r = __dev_status(md, param);
+
+	dm_put(md);
+	return r;
+}
+
+static int do_resume(struct dm_ioctl *param)
+{
+	int r = 0;
+	struct hash_cell *hc;
+	struct mapped_device *md;
+	struct dm_table *new_map;
+
+	down_write(&_hash_lock);
+
+	hc = __find_device_hash_cell(param);
+	if (!hc) {
+		DMWARN("device doesn't appear to be in the dev hash table.");
+		up_write(&_hash_lock);
+		return -ENXIO;
+	}
+
+	md = hc->md;
+	dm_get(md);
+
+	new_map = hc->new_map;
+	hc->new_map = NULL;
+	param->flags &= ~DM_INACTIVE_PRESENT_FLAG;
+
+	up_write(&_hash_lock);
+
+	/* Do we need to load a new map ? */
+	if (new_map) {
+		/* Suspend if it isn't already suspended */
+		if (!dm_suspended(md))
+			dm_suspend(md);
+
+		r = dm_swap_table(md, new_map);
+		if (r) {
+			dm_put(md);
+			dm_table_put(new_map);
+			return r;
+		}
+
+		if (dm_table_get_mode(new_map) & FMODE_WRITE)
+			set_device_ro(dm_kdev(md), 0);
+		else
+			set_device_ro(dm_kdev(md), 1);
+
+		dm_table_put(new_map);
+	}
+
+	if (dm_suspended(md))
+		r = dm_resume(md);
+
+	if (!r)
+		r = __dev_status(md, param);
+
+	dm_put(md);
+	return r;
+}
+
+/*
+ * Set or unset the suspension state of a device.
+ * If the device already is in the requested state we just return its status.
+ */
+static int dev_suspend(struct dm_ioctl *param, size_t param_size)
+{
+	if (param->flags & DM_SUSPEND_FLAG)
+		return do_suspend(param);
+
+	return do_resume(param);
 }
 
 /*
  * Copies device info back to user space, used by
  * the create and info ioctls.
  */
-static int info(struct dm_ioctl *param, struct dm_ioctl *user)
+static int dev_status(struct dm_ioctl *param, size_t param_size)
 {
+	int r;
 	struct mapped_device *md;
 
-	param->flags = 0;
-
 	md = find_device(param);
 	if (!md)
-		/*
-		 * Device not found - returns cleared exists flag.
-		 */
-		goto out;
+		return -ENXIO;
 
-	__info(md, param);
+	r = __dev_status(md, param);
 	dm_put(md);
-
-      out:
-	return results_to_user(user, param, NULL, 0);
+	return r;
 }
 
 static inline int get_mode(struct dm_ioctl *param)
@@ -547,243 +708,142 @@ static inline int get_mode(struct dm_ioctl *param)
 	return mode;
 }
 
-static int check_name(const char *name)
+static int next_target(struct dm_target_spec *last, uint32_t next, void *end,
+		       struct dm_target_spec **spec, char **target_params)
 {
-	if (strchr(name, '/')) {
-		DMWARN("invalid device name");
+	*spec = (struct dm_target_spec *) ((unsigned char *) last + next);
+	*target_params = (char *) (*spec + 1);
+
+	if (*spec < (last + 1))
+		return -EINVAL;
+
+	return invalid_str(*target_params, end);
+}
+
+static int populate_table(struct dm_table *table, struct dm_ioctl *param,
+			  size_t param_size)
+{
+	int r;
+	unsigned int i = 0;
+	struct dm_target_spec *spec = (struct dm_target_spec *) param;
+	uint32_t next = param->data_start;
+	void *end = (void *) param + param_size;
+	char *target_params;
+
+	if (!param->target_count) {
+		DMWARN("populate_table: no targets specified");
 		return -EINVAL;
 	}
 
-	return 0;
+	for (i = 0; i < param->target_count; i++) {
+
+		r = next_target(spec, next, end, &spec, &target_params);
+		if (r) {
+			DMWARN("unable to find target");
+			return r;
+		}
+
+		r = dm_table_add_target(table, spec->target_type,
+					(sector_t) spec->sector_start,
+					(sector_t) spec->length,
+					target_params);
+		if (r) {
+			DMWARN("error adding target to table");
+			return r;
+		}
+
+		next = spec->next;
+	}
+
+	return dm_table_complete(table);
 }
 
-static int create(struct dm_ioctl *param, struct dm_ioctl *user)
+static int table_load(struct dm_ioctl *param, size_t param_size)
 {
 	int r;
-	kdev_t dev;
+	struct hash_cell *hc;
 	struct dm_table *t;
-	struct mapped_device *md;
-	int minor;
-
-	r = check_name(param->name);
-	if (r)
-		return r;
 
 	r = dm_table_create(&t, get_mode(param));
 	if (r)
 		return r;
 
-	r = populate_table(t, param);
+	r = populate_table(t, param, param_size);
 	if (r) {
 		dm_table_put(t);
 		return r;
 	}
 
-	minor = (param->flags & DM_PERSISTENT_DEV_FLAG) ?
-	    minor(to_kdev_t(param->dev)) : -1;
-
-	r = dm_create(minor, t, &md);
-	if (r) {
-		dm_table_put(t);
-		return r;
-	}
-	dm_table_put(t);	/* md will have grabbed its own reference */
-
-	dev = dm_kdev(md);
-	set_device_ro(dev, (param->flags & DM_READONLY_FLAG));
-	r = dm_hash_insert(param->name, *param->uuid ? param->uuid : NULL, md);
-	dm_put(md);
-
-	return r ? r : info(param, user);
-}
-
-/*
- * Build up the status struct for each target
- */
-static int __status(struct mapped_device *md, struct dm_ioctl *param,
-		    char *outbuf, int *len)
-{
-	int i, num_targets;
-	struct dm_target_spec *spec;
-	char *outptr;
-	status_type_t type;
-	struct dm_table *table = dm_get_table(md);
-
-	if (param->flags & DM_STATUS_TABLE_FLAG)
-		type = STATUSTYPE_TABLE;
-	else
-		type = STATUSTYPE_INFO;
-
-	outptr = outbuf;
-
-	/* Get all the target info */
-	num_targets = dm_table_get_num_targets(table);
-	for (i = 0; i < num_targets; i++) {
-		struct dm_target *ti = dm_table_get_target(table, i);
-
-		if (outptr - outbuf +
-		    sizeof(struct dm_target_spec) > param->data_size) {
-			dm_table_put(table);
-			return -ENOMEM;
-		}
-
-		spec = (struct dm_target_spec *) outptr;
-
-		spec->status = 0;
-		spec->sector_start = ti->begin;
-		spec->length = ti->len;
-		strncpy(spec->target_type, ti->type->name,
-			sizeof(spec->target_type));
-
-		outptr += sizeof(struct dm_target_spec);
-
-		/* Get the status/table string from the target driver */
-		if (ti->type->status)
-			ti->type->status(ti, type, outptr,
-					 outbuf + param->data_size - outptr);
-		else
-			outptr[0] = '\0';
-
-		outptr += strlen(outptr) + 1;
-		_align(outptr, ALIGNMENT);
-		spec->next = outptr - outbuf;
+	down_write(&_hash_lock);
+	hc = __find_device_hash_cell(param);
+	if (!hc) {
+		DMWARN("device doesn't appear to be in the dev hash table.");
+		up_write(&_hash_lock);
+		return -ENXIO;
 	}
 
-	param->target_count = num_targets;
-	*len = outptr - outbuf;
-	dm_table_put(table);
+	if (hc->new_map)
+		dm_table_put(hc->new_map);
+	hc->new_map = t;
+	param->flags |= DM_INACTIVE_PRESENT_FLAG;
 
-	return 0;
+	r = __dev_status(hc->md, param);
+	up_write(&_hash_lock);
+	return r;
 }
 
-/*
- * Return the status of a device as a text string for each
- * target.
- */
-static int get_status(struct dm_ioctl *param, struct dm_ioctl *user)
+static int table_clear(struct dm_ioctl *param, size_t param_size)
 {
-	struct mapped_device *md;
-	int len = 0;
-	int ret;
-	char *outbuf = NULL;
+	int r;
+	struct hash_cell *hc;
 
-	md = find_device(param);
-	if (!md)
-		/*
-		 * Device not found - returns cleared exists flag.
-		 */
-		goto out;
+	down_write(&_hash_lock);
 
-	/* We haven't a clue how long the resultant data will be so
-	   just allocate as much as userland has allowed us and make sure
-	   we don't overun it */
-	outbuf = kmalloc(param->data_size, GFP_KERNEL);
-	if (!outbuf)
-		goto out;
-	/*
-	 * Get the status of all targets
-	 */
-	__status(md, param, outbuf, &len);
+	hc = __find_device_hash_cell(param);
+	if (!hc) {
+		DMWARN("device doesn't appear to be in the dev hash table.");
+		up_write(&_hash_lock);
+		return -ENXIO;
+	}
 
-	/*
-	 * Setup the basic dm_ioctl structure.
-	 */
-	__info(md, param);
+	if (hc->new_map) {
+		dm_table_put(hc->new_map);
+		hc->new_map = NULL;
+	}
 
-      out:
-	if (md)
-		dm_put(md);
+	param->flags &= ~DM_INACTIVE_PRESENT_FLAG;
 
-	ret = results_to_user(user, param, outbuf, len);
-
-	if (outbuf)
-		kfree(outbuf);
-
-	return ret;
-}
-
-/*
- * Wait for a device to report an event
- */
-static int wait_device_event(struct dm_ioctl *param, struct dm_ioctl *user)
-{
-	struct mapped_device *md;
-	struct dm_table *table;
-	DECLARE_WAITQUEUE(wq, current);
-
-	md = find_device(param);
-	if (!md)
-		/*
-		 * Device not found - returns cleared exists flag.
-		 */
-		goto out;
-
-	/*
-	 * Setup the basic dm_ioctl structure.
-	 */
-	__info(md, param);
-
-	/*
-	 * Wait for a notification event
-	 */
-	set_current_state(TASK_INTERRUPTIBLE);
-	table = dm_get_table(md);
-	dm_table_add_wait_queue(table, &wq);
-	dm_table_put(table);
-	dm_put(md);
-
-	yield();
-	set_current_state(TASK_RUNNING);
-
-      out:
-	return results_to_user(user, param, NULL, 0);
+	r = __dev_status(hc->md, param);
+	up_write(&_hash_lock);
+	return r;
 }
 
 /*
  * Retrieves a list of devices used by a particular dm device.
  */
-static int dep(struct dm_ioctl *param, struct dm_ioctl *user)
+static void retrieve_deps(struct dm_table *table, struct dm_ioctl *param,
+			  size_t param_size)
 {
-	int count, r;
-	struct mapped_device *md;
+	unsigned int count = 0;
 	struct list_head *tmp;
-	size_t len = 0;
-	struct dm_target_deps *deps = NULL;
-	struct dm_table *table;
+	size_t len, needed;
+	struct dm_target_deps *deps;
 
-	md = find_device(param);
-	if (!md)
-		goto out;
-	table = dm_get_table(md);
-
-	/*
-	 * Setup the basic dm_ioctl structure.
-	 */
-	__info(md, param);
+	deps = get_result_buffer(param, param_size, &len);
 
 	/*
 	 * Count the devices.
 	 */
-	count = 0;
 	list_for_each(tmp, dm_table_get_devices(table))
-	    count++;
+		count++;
 
 	/*
-	 * Allocate a kernel space version of the dm_target_status
-	 * struct.
+	 * Check we have enough space.
 	 */
-	if (array_too_big(sizeof(*deps), sizeof(*deps->dev), count)) {
-		dm_table_put(table);
-		dm_put(md);
-		return -ENOMEM;
-	}
-
-	len = sizeof(*deps) + (sizeof(*deps->dev) * count);
-	deps = kmalloc(len, GFP_KERNEL);
-	if (!deps) {
-		dm_table_put(table);
-		dm_put(md);
-		return -ENOMEM;
+	needed = sizeof(*deps) + (sizeof(*deps->dev) * count);
+	if (len < needed) {
+		param->flags |= DM_BUFFER_FULL_FLAG;
+		return;
 	}
 
 	/*
@@ -795,134 +855,171 @@ static int dep(struct dm_ioctl *param, struct dm_ioctl *user)
 		struct dm_dev *dd = list_entry(tmp, struct dm_dev, list);
 		deps->dev[count++] = dd->bdev->bd_dev;
 	}
-	dm_table_put(table);
-	dm_put(md);
 
-      out:
-	r = results_to_user(user, param, deps, len);
-
-	kfree(deps);
-	return r;
+	param->data_size = param->data_start + needed;
 }
 
-static int remove(struct dm_ioctl *param, struct dm_ioctl *user)
-{
-	struct hash_cell *hc;
-
-	down_write(&_hash_lock);
-	hc = *param->uuid ? __get_uuid_cell(param->uuid) :
-	    __get_name_cell(param->name);
-	if (!hc) {
-		DMWARN("device doesn't appear to be in the dev hash table.");
-		up_write(&_hash_lock);
-		return -EINVAL;
-	}
-
-	/*
-	 * You may ask the interface to drop its reference to an
-	 * in use device.  This is no different to unlinking a
-	 * file that someone still has open.  The device will not
-	 * actually be destroyed until the last opener closes it.
-	 * The name and uuid of the device (both are interface
-	 * properties) will be available for reuse immediately.
-	 *
-	 * You don't want to drop a _suspended_ device from the
-	 * interface, since that will leave you with no way of
-	 * resuming it.
-	 */
-	if (dm_suspended(hc->md)) {
-		DMWARN("refusing to remove a suspended device.");
-		up_write(&_hash_lock);
-		return -EPERM;
-	}
-
-	__hash_remove(hc);
-	up_write(&_hash_lock);
-	return 0;
-}
-
-static int remove_all(struct dm_ioctl *param, struct dm_ioctl *user)
-{
-	dm_hash_remove_all();
-	return 0;
-}
-
-static int suspend(struct dm_ioctl *param, struct dm_ioctl *user)
+static int table_deps(struct dm_ioctl *param, size_t param_size)
 {
 	int r;
 	struct mapped_device *md;
+	struct dm_table *table;
 
 	md = find_device(param);
 	if (!md)
 		return -ENXIO;
 
-	if (param->flags & DM_SUSPEND_FLAG)
-		r = dm_suspend(md);
-	else
-		r = dm_resume(md);
+	r = __dev_status(md, param);
+	if (r)
+		goto out;
 
+	table = dm_get_table(md);
+	if (table) {
+		retrieve_deps(table, param, param_size);
+		dm_table_put(table);
+	}
+
+ out:
 	dm_put(md);
 	return r;
 }
 
-static int reload(struct dm_ioctl *param, struct dm_ioctl *user)
+/*
+ * Build up the status struct for each target
+ */
+static void retrieve_status(struct dm_table *table, struct dm_ioctl *param,
+			    size_t param_size)
+{
+	unsigned int i, num_targets;
+	struct dm_target_spec *spec;
+	char *outbuf, *outptr;
+	status_type_t type;
+	size_t remaining, len, used = 0;
+
+	outptr = outbuf = get_result_buffer(param, param_size, &len);
+
+	if (param->flags & DM_STATUS_TABLE_FLAG)
+		type = STATUSTYPE_TABLE;
+	else
+		type = STATUSTYPE_INFO;
+
+	/* Get all the target info */
+	num_targets = dm_table_get_num_targets(table);
+	for (i = 0; i < num_targets; i++) {
+		struct dm_target *ti = dm_table_get_target(table, i);
+
+		remaining = len - (outptr - outbuf);
+		if (remaining < sizeof(struct dm_target_spec)) {
+			param->flags |= DM_BUFFER_FULL_FLAG;
+			break;
+		}
+
+		spec = (struct dm_target_spec *) outptr;
+
+		spec->status = 0;
+		spec->sector_start = ti->begin;
+		spec->length = ti->len;
+		strncpy(spec->target_type, ti->type->name,
+			sizeof(spec->target_type));
+
+		outptr += sizeof(struct dm_target_spec);
+		remaining = len - (outptr - outbuf);
+
+		/* Get the status/table string from the target driver */
+		if (ti->type->status) {
+			if (ti->type->status(ti, type, outptr, remaining)) {
+				param->flags |= DM_BUFFER_FULL_FLAG;
+				break;
+			}
+		} else
+			outptr[0] = '\0';
+
+		outptr += strlen(outptr) + 1;
+		used = param->data_start + (outptr - outbuf);
+
+		align_ptr(outptr);
+		spec->next = outptr - outbuf;
+	}
+
+	if (used)
+		param->data_size = used;
+
+	param->target_count = num_targets;
+}
+
+/*
+ * Return the status of a device as a text string for each
+ * target.
+ */
+static int table_status(struct dm_ioctl *param, size_t param_size)
 {
 	int r;
-	kdev_t dev;
 	struct mapped_device *md;
-	struct dm_table *t;
-
-	r = dm_table_create(&t, get_mode(param));
-	if (r)
-		return r;
-
-	r = populate_table(t, param);
-	if (r) {
-		dm_table_put(t);
-		return r;
-	}
+	struct dm_table *table;
 
 	md = find_device(param);
-	if (!md) {
-		dm_table_put(t);
+	if (!md)
 		return -ENXIO;
+
+	r = __dev_status(md, param);
+	if (r)
+		goto out;
+ 
+	table = dm_get_table(md);
+	if (table) {
+		retrieve_status(table, param, param_size);
+		dm_table_put(table);
 	}
 
-	r = dm_swap_table(md, t);
-	if (r) {
-		dm_put(md);
-		dm_table_put(t);
-		return r;
-	}
-	dm_table_put(t);	/* md will have taken its own reference */
-
-	dev = dm_kdev(md);
-	set_device_ro(dev, (param->flags & DM_READONLY_FLAG));
+ out:
 	dm_put(md);
-
-	r = info(param, user);
 	return r;
 }
 
-static int rename(struct dm_ioctl *param, struct dm_ioctl *user)
+/*
+ * Wait for a device to report an event
+ */
+static int dev_wait(struct dm_ioctl *param, size_t param_size)
 {
 	int r;
-	char *new_name = (char *) param + param->data_start;
+	struct mapped_device *md;
+	struct dm_table *table;
+	DECLARE_WAITQUEUE(wq, current);
 
-	if (valid_str(new_name, (void *) param,
-		      (void *) param + param->data_size)) {
-		DMWARN("Invalid new logical volume name supplied.");
-		return -EINVAL;
+	md = find_device(param);
+	if (!md)
+		return -ENXIO;
+
+	/*
+	 * Wait for a notification event
+	 */
+	set_current_state(TASK_INTERRUPTIBLE);
+	if (!dm_add_wait_queue(md, &wq, param->event_nr)) {
+		schedule();
+		dm_remove_wait_queue(md, &wq);
+	}
+	set_current_state(TASK_RUNNING);
+
+	/*
+	 * The userland program is going to want to know what
+	 * changed to trigger the event, so we may as well tell
+	 * him and save an ioctl.
+	 */
+	r = __dev_status(md, param);
+	if (r)
+		goto out;
+
+	table = dm_get_table(md);
+	if (table) {
+		retrieve_status(table, param, param_size);
+		dm_table_put(table);
 	}
 
-	r = check_name(new_name);
-	if (r)
-		return r;
-
-	return dm_hash_rename(param->name, new_name);
+ out:
+	dm_put(md);
+	return r;
 }
 
-
 /*-----------------------------------------------------------------
  * Implementation of open/close/ioctl on the special char
  * device.
@@ -935,15 +1032,19 @@ static ioctl_fn lookup_ioctl(unsigned int cmd)
 	} _ioctls[] = {
 		{DM_VERSION_CMD, NULL},	/* version is dealt with elsewhere */
 		{DM_REMOVE_ALL_CMD, remove_all},
-		{DM_DEV_CREATE_CMD, create},
-		{DM_DEV_REMOVE_CMD, remove},
-		{DM_DEV_RELOAD_CMD, reload},
-		{DM_DEV_RENAME_CMD, rename},
-		{DM_DEV_SUSPEND_CMD, suspend},
-		{DM_DEV_DEPS_CMD, dep},
-		{DM_DEV_STATUS_CMD, info},
-		{DM_TARGET_STATUS_CMD, get_status},
-		{DM_TARGET_WAIT_CMD, wait_device_event},
+		{DM_LIST_DEVICES_CMD, list_devices},
+
+		{DM_DEV_CREATE_CMD, dev_create},
+		{DM_DEV_REMOVE_CMD, dev_remove},
+		{DM_DEV_RENAME_CMD, dev_rename},
+		{DM_DEV_SUSPEND_CMD, dev_suspend},
+		{DM_DEV_STATUS_CMD, dev_status},
+		{DM_DEV_WAIT_CMD, dev_wait},
+
+		{DM_TABLE_LOAD_CMD, table_load},
+		{DM_TABLE_CLEAR_CMD, table_clear},
+		{DM_TABLE_DEPS_CMD, table_deps},
+		{DM_TABLE_STATUS_CMD, table_status}
 	};
 
 	return (cmd >= ARRAY_SIZE(_ioctls)) ? NULL : _ioctls[cmd].fn;
@@ -953,7 +1054,7 @@ static ioctl_fn lookup_ioctl(unsigned int cmd)
  * As well as checking the version compatibility this always
  * copies the kernel interface version out.
  */
-static int check_version(int cmd, struct dm_ioctl *user)
+static int check_version(unsigned int cmd, struct dm_ioctl *user)
 {
 	uint32_t version[3];
 	int r = 0;
@@ -1013,15 +1114,19 @@ static int copy_params(struct dm_ioctl *user, struct dm_ioctl **param)
 
 static int validate_params(uint cmd, struct dm_ioctl *param)
 {
+	/* Always clear this flag */
+	param->flags &= ~DM_BUFFER_FULL_FLAG;
+
 	/* Ignores parameters */
-	if (cmd == DM_REMOVE_ALL_CMD)
+	if (cmd == DM_REMOVE_ALL_CMD || cmd == DM_LIST_DEVICES_CMD)
 		return 0;
 
-	/* Unless creating, either name of uuid but not both */
+	/* Unless creating, either name or uuid but not both */
 	if (cmd != DM_DEV_CREATE_CMD) {
 		if ((!*param->uuid && !*param->name) ||
 		    (*param->uuid && *param->name)) {
-			DMWARN("one of name or uuid must be supplied");
+			DMWARN("one of name or uuid must be supplied, cmd(%u)",
+			       cmd);
 			return -EINVAL;
 		}
 	}
@@ -1036,10 +1141,12 @@ static int validate_params(uint cmd, struct dm_ioctl *param)
 static int ctl_ioctl(struct inode *inode, struct file *file,
 		     uint command, ulong u)
 {
-	int r = 0, cmd;
+	int r = 0;
+	unsigned int cmd;
 	struct dm_ioctl *param;
 	struct dm_ioctl *user = (struct dm_ioctl *) u;
 	ioctl_fn fn = NULL;
+	size_t param_size;
 
 	/* only root can play with this */
 	if (!capable(CAP_SYS_ADMIN))
@@ -1070,21 +1177,38 @@ static int ctl_ioctl(struct inode *inode, struct file *file,
 		return -ENOTTY;
 	}
 
+	/*
+	 * FIXME: I don't like this, we're trying to avoid low
+	 * memory issues when a device is suspended.
+	 */
+	current->flags |= PF_MEMALLOC;
+
 	/*
 	 * Copy the parameters into kernel space.
 	 */
 	r = copy_params(user, &param);
-	if (r)
-		return r;
-
-	r = validate_params(cmd, param);
 	if (r) {
-		free_params(param);
+		current->flags &= ~PF_MEMALLOC;
 		return r;
 	}
 
-	r = fn(param, user);
+	r = validate_params(cmd, param);
+	if (r)
+		goto out;
+
+	param_size = param->data_size;
+	param->data_size = sizeof(*param);
+	r = fn(param, param_size);
+
+	/*
+	 * Copy the results back to userland.
+	 */
+	if (!r && copy_to_user(user, param, param->data_size))
+		r = -EFAULT;
+
+ out:
 	free_params(param);
+	current->flags &= ~PF_MEMALLOC;
 	return r;
 }
 
